################################################################
# Library
################################################################
import cv2
import os
import time
import numpy as np
import socket

import firebase_admin
from firebase_admin import credentials
from firebase_admin import db
from google.cloud import storage

################################################################
# Color
################################################################

black = (0,0,0)
blue = (255,0,0)
green = (0,255,0)
red = (0,0,255)

################################################################
# Function
################################################################

# Rotate opencv image
def cv2_rotate(img, angle):
    for _ in range(angle // 90):
        img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)
    return img

# Pre-Processing
def preprocessing(img):
    # Conver HSV and masking
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    mask = cv2.inRange(hsv, (0, 0, 230), (180, 255,255)) > 0
    preproc = np.full_like(img, 255)
    preproc[mask] = 0
    # Return image
    return preproc

# Crop image
def crop(img, preproc, scale, pos):
    # Calculate coordination
    x, x_t = pos[0], int(img.shape[1] / scale)
    y, y_t = pos[1], x_t // 3
    # Crop and draw bounding box
    crop = preproc[y:y+y_t, x:x+x_t]
    bbox = cv2.rectangle(img, (x, y), (x+x_t, y+y_t), green, 3)
    return crop, bbox

# Boxing
def boxing(img, scale, pos):
    # Calculate coordination
    x, x_t = pos[0], int(img.shape[1] / scale)
    y, y_t = pos[1], x_t // 3
    # Crop and draw bounding box
    bbox = cv2.rectangle(img, (x, y), (x+x_t, y+y_t), green, 3)
    cv2.imwrite("/tmp/digital.png", bbox)
    return bbox, "/tmp/digital.png"

# Draw coordinate
def coordinate(img):
    # Draw coordinate
    origin = (0, 0)
    x_end = (0, img.shape[0])
    y_end = (img.shape[1], 0)
    img = cv2.line(img, origin, x_end, blue, 3)
    img = cv2.line(img, origin, y_end, blue, 3)
    # Graduation
    for x in range(0, img.shape[1], 50):
        x1, x2, x3 = (x, 0), (x, 10), (x, 20)
        img = cv2.line(img, x1, x2, blue, 3)
        img = cv2.putText(img, str(x), x3, cv2.FONT_HERSHEY_SIMPLEX, 0.5, black, 1, cv2.LINE_AA)
    for y in range(0, img.shape[0], 50):
        y1, y2, y3 = (0, y), (10, y), (20, y)
        img = cv2.line(img, y1, y2, blue, 3)
        img = cv2.putText(img, str(y), y3, cv2.FONT_HERSHEY_SIMPLEX, 0.5, black, 1, cv2.LINE_AA)

    cv2.imwrite("/tmp/digital.png", img)
    return img, "/tmp/digital.png"

# Find Most frequent value
def most_frequent(l):
    return max(set(l), key=l.count)

# Save Pre-Processing
def cv2_save(cropframe, bbox, preproc=None, debugframe=None, debug_mode=False):
    # Save Frame PNG file
    if debug_mode:
        cv2.imwrite("image/preproc.png", preproc)
        cv2.imwrite("image/cropframe.png", cropframe)
        cv2.imwrite("image/debugframe{:d}.png".format(bbox), debugframe)
        cv2.imshow("debugframe",debugframe)
        return "image/cropframe.png", "image/debugframe{:d}.png".format(bbox)
    else:
        cv2.imwrite("/tmp/cropframe.png", cropframe)
        cv2.imwrite("/tmp/debugframe{:d}.png".format(bbox), debugframe)
        return "/tmp/cropframe.png", "/tmp/debugframe{:d}.png".format(bbox)

# Get Firebase Information
def firebase_info(info_pth):
    # Read JSON file
    key = os.path.join(info_pth, "key.json")
    # Read URL
    with open(os.path.join(info_pth, "url"), "r") as f:
        url = [line[:-1] for line in f]
    return key, url

def get_ipaddr():
    # Set socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # Get ip address by sending socket
    try:       
        sock.connect(('10.255.255.255', 1))
        ipaddr = sock.getsockname()[0]
    except Exception:
        ipaddr = '127.0.0.1'
    finally:
        sock.close()
    return ipaddr

################################################################
# Class
################################################################

class firebase:
    # Config
    ELAPSED_TIME = 10
    LIMIT_DATA = 10000

    # Constructor
    def __init__(self, device, key, url, idx):
        # Initialize Firebase
        if idx == 0:
            cred = credentials.Certificate(key)
            firebase_admin.initialize_app(cred, {
                'databaseURL' : url
            })
        self.device = device
        self.database = db.reference()
        # Initialize list
        self.init = time.time()
        self.data_list = []
        self.data_idx = 0
        # Clean Database
        if idx == 0:
            self.clean(key=self.device)
        # Update IP addresss
        if idx == 0:
            self.update_ip()

    # Update Data
    def update(self, bbox, data):
        # Counte elapsed time
        elasped = int(time.time() - self.init)
        # Append digit data
        if data.isdigit():
            self.data_list.append(int(data))
            print("[Info ] Bbox[{:02d}] : {}".format(bbox, data))
        else:
            return
        # Store the data if list is full
        if elasped >= firebase.ELAPSED_TIME and len(self.data_list) >= 10:
            digit = most_frequent(self.data_list)
            self.store(bbox=bbox, data=digit)
            self.data_list = []
            self.init = time.time()

    # Store Information
    def store(self, bbox, data):
        # Bbox and time index
        bid = "Box{:d}".format(bbox)
        self.data_idx += 1
        did = str(self.data_idx)
        # Check data is integer
        fdb = self.database.child(self.device)
        fdb.child(bid).update({did:data})
        print("[Info ] Send data to firebase")

    # Update IP address
    def update_ip(self):
        address =  get_ipaddr()
        self.database.child(self.device).update({"IP" : address})
        print("[Info ] Send ip address to firebase")

    # Clean Firebase
    def clean(self, key):
        json = self.database.get()
        # Clean json file
        if json is None:
            return
        else:
            if key in json.keys():
                delete_key = self.database.child(key)
                delete_key.delete()

class Storage:
    # Constructor
    def __init__(self, url):
        self.client = storage.Client()
        self.bucket = self.client.get_bucket(url)

    # Upload image to storage
    def upload_image(self, image):
        imageBlob = self.bucket.blob('digital.png')
        imageBlob.upload_from_filename(image)
        print("[Info ] Send image to firebase")
